OPENSENSEMAP_API_LIVE = "https://api.opensensemap.org/"
"""
The live OpenSeneMap API server
"""

OPENSENSEMAP_API_TEST = "https://api.testing.opensensemap.org/"
"""
The test OpenSeneMap API server
"""

SIGN_IN = "users/sign-in"
SIGN_OUT = "users/sign-out"
REFRESH = "users/refresh-auth"
USER_DETAILS = "users/me"
BOXES = "boxes"
